﻿using System;
using System.Linq;
using System.Collections.Generic;

namespace Achiral
{
    using Achiral.Extension;
    using System.Diagnostics;

    // Declaration
    public partial struct Tuple<TArg1, TArg2>
    {
        public readonly TArg1 Arg1;
        public readonly TArg2 Arg2;
        public Tuple(TArg1 arg1, TArg2 arg2)
        {
            this.Arg1 = arg1;
            this.Arg2 = arg2;
        }
    }

    public partial struct Tuple<TArg1, TArg2, TArg3>
    {
        public readonly TArg1 Arg1;
        public readonly TArg2 Arg2;
        public readonly TArg3 Arg3;
        public Tuple(TArg1 arg1, TArg2 arg2, TArg3 arg3)
        {
            this.Arg1 = arg1;
            this.Arg2 = arg2;
            this.Arg3 = arg3;
        }
    }

    public partial struct Tuple<TArg1, TArg2, TArg3, TArg4>
    {
        public readonly TArg1 Arg1;
        public readonly TArg2 Arg2;
        public readonly TArg3 Arg3;
        public readonly TArg4 Arg4;
        public Tuple(TArg1 arg1, TArg2 arg2, TArg3 arg3, TArg4 arg4)
        {
            this.Arg1 = arg1;
            this.Arg2 = arg2;
            this.Arg3 = arg3;
            this.Arg4 = arg4;
        }
    }

    // Select
    public partial struct Tuple<TArg1, TArg2>
    {
        public TResult Select<TResult>(Func<TArg1, TArg2, TResult> selector)
        {
            return selector(this.Arg1, this.Arg2);
        }
    }
    public partial struct Tuple<TArg1, TArg2, TArg3>
    {
        public TResult Select<TResult>(Func<TArg1, TArg2, TArg3, TResult> selector)
        {
            return selector(this.Arg1, this.Arg2, this.Arg3);
        }
    }
    public partial struct Tuple<TArg1, TArg2, TArg3, TArg4>
    {
        public TResult Select<TResult>(Func<TArg1, TArg2, TArg3, TArg4, TResult> selector)
        {
            return selector(this.Arg1, this.Arg2, this.Arg3, this.Arg4);
        }
    }

    // DebuggerDisplay
    [DebuggerDisplay("({Arg1}, {Arg2})")]
    public partial struct Tuple<TArg1, TArg2> { }
    [DebuggerDisplay("({Arg1}, {Arg2}, {Arg3})")]
    public partial struct Tuple<TArg1, TArg2, TArg3> { }
    [DebuggerDisplay("({Arg1}, {Arg2}, {Arg3}, {Arg4})")]
    public partial struct Tuple<TArg1, TArg2, TArg3, TArg4> { }

    // Make
    public static partial class Make
    {
        public static Tuple<TArg1, TArg2> Tuple<TArg1, TArg2>(TArg1 arg1, TArg2 arg2)
        {
            return new Tuple<TArg1, TArg2>(arg1, arg2);
        }
        public static Tuple<TArg1, TArg2, TArg3> Tuple<TArg1, TArg2, TArg3>(TArg1 arg1, TArg2 arg2, TArg3 arg3)
        {
            return new Tuple<TArg1, TArg2, TArg3>(arg1, arg2, arg3);
        }
        public static Tuple<TArg1, TArg2, TArg3, TArg4> Tuple<TArg1, TArg2, TArg3, TArg4>(TArg1 arg1, TArg2 arg2, TArg3 arg3, TArg4 arg4)
        {
            return new Tuple<TArg1, TArg2, TArg3, TArg4>(arg1, arg2, arg3, arg4);
        }
    }

    // IEquatable Implimentation (Generated by ReSharper 3.1)

    public partial struct Tuple<TArg1, TArg2> : IEquatable<Tuple<TArg1, TArg2>>
    {
        public static bool operator !=(Tuple<TArg1, TArg2> tuple1, Tuple<TArg1, TArg2> tuple2)
        {
            return !tuple1.Equals(tuple2);
        }

        public static bool operator ==(Tuple<TArg1, TArg2> tuple1, Tuple<TArg1, TArg2> tuple2)
        {
            return tuple1.Equals(tuple2);
        }

        public bool Equals(Tuple<TArg1, TArg2> tuple)
        {
            return Equals(Arg1, tuple.Arg1) && Equals(Arg2, tuple.Arg2);
        }

        public override bool Equals(object obj)
        {
            if (!(obj is Tuple<TArg1, TArg2>)) return false;
            return Equals((Tuple<TArg1, TArg2>)obj);
        }

        public override int GetHashCode()
        {
            return (Arg1 != null ? Arg1.GetHashCode() : 0) + 29 * (Arg2 != null ? Arg2.GetHashCode() : 0);
        }
    }


    public partial struct Tuple<TArg1, TArg2, TArg3> : IEquatable<Tuple<TArg1, TArg2, TArg3>>
    {
        public static bool operator !=(Tuple<TArg1, TArg2, TArg3> tuple1, Tuple<TArg1, TArg2, TArg3> tuple2)
        {
            return !tuple1.Equals(tuple2);
        }

        public static bool operator ==(Tuple<TArg1, TArg2, TArg3> tuple1, Tuple<TArg1, TArg2, TArg3> tuple2)
        {
            return tuple1.Equals(tuple2);
        }

        public bool Equals(Tuple<TArg1, TArg2, TArg3> tuple)
        {
            if (!Make.Equality(Arg1, tuple.Arg1)) return false;
            if (!Make.Equality(Arg2, tuple.Arg2)) return false;
            if (!Make.Equality(Arg3, tuple.Arg3)) return false;
            return true;
        }

        public override bool Equals(object obj)
        {
            if (!(obj is Tuple<TArg1, TArg2, TArg3>)) return false;
            return Equals((Tuple<TArg1, TArg2, TArg3>)obj);
        }

        public override int GetHashCode()
        {
            int result = Arg1 != null ? Arg1.GetHashCode() : 0;
            result = 29 * result + (Arg2 != null ? Arg2.GetHashCode() : 0);
            result = 29 * result + (Arg3 != null ? Arg3.GetHashCode() : 0);
            return result;
        }
    }

    public partial struct Tuple<TArg1, TArg2, TArg3, TArg4> : IEquatable<Tuple<TArg1, TArg2, TArg3, TArg4>>
    {
        public static bool operator !=(Tuple<TArg1, TArg2, TArg3, TArg4> tuple1, Tuple<TArg1, TArg2, TArg3, TArg4> tuple2)
        {
            return !tuple1.Equals(tuple2);
        }

        public static bool operator ==(Tuple<TArg1, TArg2, TArg3, TArg4> tuple1, Tuple<TArg1, TArg2, TArg3, TArg4> tuple2)
        {
            return tuple1.Equals(tuple2);
        }

        public bool Equals(Tuple<TArg1, TArg2, TArg3, TArg4> tuple)
        {
            if (!Make.Equality(Arg1, tuple.Arg1)) return false;
            if (!Make.Equality(Arg2, tuple.Arg2)) return false;
            if (!Make.Equality(Arg3, tuple.Arg3)) return false;
            if (!Make.Equality(Arg4, tuple.Arg4)) return false;
            return true;
        }

        public override bool Equals(object obj)
        {
            if (!(obj is Tuple<TArg1, TArg2, TArg3, TArg4>)) return false;
            return Equals((Tuple<TArg1, TArg2, TArg3, TArg4>)obj);
        }

        public override int GetHashCode()
        {
            int result = Arg1 != null ? Arg1.GetHashCode() : 0;
            result = 29 * result + (Arg2 != null ? Arg2.GetHashCode() : 0);
            result = 29 * result + (Arg3 != null ? Arg3.GetHashCode() : 0);
            result = 29 * result + (Arg4 != null ? Arg4.GetHashCode() : 0);
            return result;
        }
    }
}